# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2010, 2011, 2012,
# 2013, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.


class InvenioWebSearchUnknownCollectionError(Exception):
    """Exception for bad collection."""

    def __init__(self, colname):
        """Initialisation."""
        self.colname = colname

    def __str__(self):
        """String representation."""
        return repr(self.colname)


class InvenioWebSearchWildcardLimitError(Exception):
    """Exception raised when query limit reached."""

    def __init__(self, res):
        """Initialization."""
        self.res = res

class InvenioWebSearchReferstoLimitError(Exception):
    """Raise when CFG_WEBSEARCH_MAX_RECORDS_REFERSTO limit is reached."""

    def __init__(self, res):
        """Initialization."""
        self.res = res

class InvenioWebSearchCitedbyLimitError(Exception):
    """Raise when CFG_WEBSEARCH_MAX_RECORDS_CITEDBY limit is reached."""

    def __init__(self, res):
        """Initialization."""
        self.res = res



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
"""
WebSearch service to do some basic math
"""
import re
import cgi
from invenio.config import CFG_SITE_LANG
from invenio_search.services import SearchService

numerics_and_operators_re = re.compile('[^\s\d\+\-\*/\.\(\)]')
integers_re = re.compile('([^\.\d]*|^)(\d*(\.)?\d*)([^\.\d]*|$)')

__plugin_version__ = "Search Service Plugin API 1.0"

class MathCalculatorService(SearchService):
    """
    Do basic math
    """

    def get_description(self, ln=CFG_SITE_LANG):
        "Return service description"
        return "Return evaluated math expression"

    def answer(self, req, user_info, of, cc, colls_to_search, p, f, search_units, ln):
        """
        Answer question given by context.

        Return (relevance, html_string) where relevance is integer
        from 0 to 100 indicating how relevant to the question the
        answer is (see C{CFG_WEBSEARCH_SERVICE_MAX_SERVICE_ANSWER_RELEVANCE} for details) ,
        and html_string being a formatted answer.
        """
        if not p.strip():
            return (0, '')
        expression = numerics_and_operators_re.sub("", p)
        if expression == p:
            # That is safe expression

            def make_float(matchobj):
                prefix =  matchobj.group(1) or ''
                suffix = matchobj.group(4) or ''
                middle = matchobj.group(2) or ''
                if middle and not '.' in middle:
                    middle += '.0'
                return prefix + middle + suffix

            expression = integers_re.sub(make_float, expression)
            try:
                result = eval(expression)
                if result.is_integer():
                    result = int(result)
                return (100, '<strong>' + cgi.escape(p.strip()) + ' = '  + \
                        cgi.escape(str(result)) + \
                        '</strong>')
            except Exception, e:
                return (0, '')
        else:
            return (0, '')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Add a table for facets configuration."""

from __future__ import print_function

import warnings

from invenio.ext.sqlalchemy import db
from invenio.modules.upgrader.api import op
from sqlalchemy.dialects import mysql

depends_on = ['invenio_release_1_1_0']


def info():
    """Show module info message."""
    return "Adds the table with facets configuration."


def do_upgrade():
    """Add the table with facets configuration."""
    if not op.has_table('facet_collection'):
        op.create_table(
            'facet_collection',
            db.Column('id', mysql.INTEGER(), nullable=False),
            db.Column('id_collection', mysql.INTEGER(), nullable=False),
            db.Column('order', mysql.INTEGER(), nullable=False),
            db.Column('facet_name', db.String(length=80), nullable=False),
            db.ForeignKeyConstraint(['id_collection'], ['collection.id'], ),
            db.PrimaryKeyConstraint('id'),
            mysql_charset='utf8',
            mysql_engine='MyISAM'
        )
    else:
        warnings.warn("*** Creation of table 'facet_collection' skipped!")


def post_upgrade():
    """Facet configuration info."""
    print('NOTE: You need to configure facets to have them shown using\n'
          'flask-admin module at /admin/facetcollection. \n\n'
          'Adding them to the default collection with id 1 makes them the\n'
          'default set. The default set is shown for every collection which\n'
          'does not have facets set configured')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Flask-Admin page to configure facets sets per collection."""

from __future__ import unicode_literals

from invenio.ext.admin.views import ModelView
from invenio.ext.sqlalchemy import db

from werkzeug.local import LocalProxy

from wtforms.fields import IntegerField, SelectField
from wtforms.validators import DataRequired, ValidationError

from .registry import facets


def is_place_taken(form, field):
    """Check if the given place for given collection is already taken.

    :param form: the form containing the validated field
    :param field: the validated field
    """
    from invenio_collections.models import FacetCollection
    order = field.data
    collection = form.data['collection']
    if not collection:
        return

    if FacetCollection.is_place_taken(collection.id, order):
        raise ValidationError('A facet on the given place already exists')


def is_duplicated(form, field):
    """Check if the given facet is already assigned to this collection.

    :param form: the form containing the validated field
    :param field: the validated field
    """
    from invenio_collections.models import FacetCollection
    facet_name = field.data
    collection = form.data['collection']
    if not collection:
        return

    if FacetCollection.is_duplicated(collection.id, facet_name):
        raise ValidationError(
            'This facet is already assigned to this collection.')


def is_module_facet_module(form, field):
    """Check if the given module is a proper one.

    :param form: the form containing the validated field
    :param field: the validated field
    """
    facet_name = field.data

    if facets.get(facet_name) is None:
        raise ValidationError('The given facet does not exist.')


def collection_query_factory():
    """Return all collections."""
    from invenio_collections.models import Collection
    return db.session.query(Collection).order_by(Collection.id),


class FacetsAdmin(ModelView):

    """Flask-Admin module to manage facets configuration."""

    _can_create = True
    _can_edit = True
    _can_delete = True

    column_list = (
        'collection', 'order', 'facet_name',
    )

    form_args = {
        'collection': {
            'validators': [
                DataRequired(),
            ],
            'allow_blank': False,
            'query_factory': collection_query_factory,
        },
        'order': {
            'validators': [
                is_place_taken,
                DataRequired(),
            ],
        },
        'facet_name': {
            'validators': [
                is_module_facet_module,
                is_duplicated,
                DataRequired(),
            ],
            'choices': LocalProxy(lambda: [
                (facet_name, facet_name) for facet_name in facets.keys()
            ]),
        },
    }

    form_overrides = {
        'facet_name': SelectField,
        'order': IntegerField,
    }

    column_default_sort = 'id_collection'

    def __init__(self, app, *args, **kwargs):
        """Constructor.

        :param app: flask application
        """
        super(FacetsAdmin, self).__init__(*args, **kwargs)


def register_admin(app, admin):
    """Called on app initialization to register administration interface."""
    from invenio_collections.models import FacetCollection
    admin.add_view(
        FacetsAdmin(app, FacetCollection, db.session, name='Facets',
                    category="", menu_icon_type='fa', menu_icon_value='fa-folder')
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implementation of search results caching."""

from intbitset import intbitset
from flask import current_app

from invenio.base.globals import cfg
from invenio.ext.cache import cache
from invenio.legacy.miscutil.data_cacher import DataCacher, DataCacherProxy
from invenio.utils.hash import md5

from .models import Field, Fieldname

search_results_cache = cache


def get_search_query_id(**kwargs):
    """Return unique query indentifier."""
    p = kwargs.get('p', '').strip()
    f = kwargs.get('f', '')
    cc = kwargs.get('cc', '')
    wl = kwargs.get('wl', '')
    so = kwargs.get('so', '')
    sf = kwargs.get('sf', '')
    return md5(repr((p, f, cc, wl, sf, so))).hexdigest()


def get_search_results_cache_key(**kwargs):
    """Return key for search results cache."""
    return cfg['CFG_SEARCH_RESULTS_CACHE_PREFIX'] + \
        get_search_query_id(**kwargs)


def get_search_results_cache_key_from_qid(qid=None):
    """Return key for search results cache from query identifier."""
    if qid is not None:
        return cfg['CFG_SEARCH_RESULTS_CACHE_PREFIX'] + qid


def get_collection_name_from_cache(qid):
    """Return collection name from query identifier."""
    try:
        return search_results_cache.get(
            get_search_results_cache_key_from_qid(qid) + '::cc')
    except Exception:
        current_app.logger.exception('Invalid collection name cache.')


def get_pattern_from_cache(qid):
    """Return pattern from query identifier."""
    try:
        return search_results_cache.get(
            get_search_results_cache_key_from_qid(qid) + '::p')
    except Exception:
        current_app.logger.exception('Invalid search pattern cache.')


def set_results_cache(results, query, collection_name=None, timeout=None):
    """Store search results in cache."""
    if cfg['CFG_WEBSEARCH_SEARCH_CACHE_SIZE'] <= 0:
        return

    timeout = timeout or cfg['CFG_WEBSEARCH_SEARCH_CACHE_TIMEOUT']
    collection_name = collection_name or cfg['CFG_SITE_NAME']
    qid = get_search_results_cache_key(p=query, cc=collection_name)

    search_results_cache.set(qid, results.fastdump(), timeout=timeout)
    search_results_cache.set(qid + '::p', query, timeout=timeout)
    search_results_cache.set(qid + '::cc', collection_name, timeout=timeout)


def get_results_cache(query, collection_name=None):
    """Get results from cache."""
    collection_name = collection_name or cfg['CFG_SITE_NAME']
    qid = get_search_results_cache_key(p=query, cc=collection_name)
    try:
        results = search_results_cache.get(qid)
        if results is not None:
            return intbitset().fastload(results)
    except Exception:
        current_app.logger.exception('Invalid search results cache.')


class FieldI18nNameDataCacher(DataCacher):

    """Provide cache for I18N field names.

    This class is not to be used directly; use function get_field_i18nname()
    instead.
    """

    def __init__(self):
        def cache_filler():
            res = Field.query.join(Fieldname).filter(
                Fieldname.type == 'ln'
            ).values(Field.name, 'ln', 'value')
            ret = {}
            for f, ln, i18nname in res:
                if i18nname:
                    if f not in ret:
                        ret[f] = {}
                    ret[f][ln] = i18nname
            return ret

        def timestamp_verifier():
            from invenio.legacy.dbquery import get_table_update_time
            return get_table_update_time('fieldname')

        DataCacher.__init__(self, cache_filler, timestamp_verifier)

field_i18nname_cache = DataCacherProxy(FieldI18nNameDataCacher)


def get_field_i18nname(f, ln=None, verify_cache_timestamp=True):
    """
    Return nicely formatted field name (of type 'ln', 'long name') for
    field F in language LN.

    If VERIFY_CACHE_TIMESTAMP is set to True, then verify DB timestamp
    and field I18N name cache timestamp and refresh cache from the DB
    if needed.  Otherwise don't bother checking DB timestamp and
    return the cached value.  (This is useful when get_field_i18nname
    is called inside a loop.)
    """
    ln = ln or cfg['CFG_SITE_LANG']
    if verify_cache_timestamp:
        field_i18nname_cache.recreate_cache_if_needed()
    out = f
    try:
        out = field_i18nname_cache.cache[f][ln]
    except KeyError:
        pass  # translation in LN does not exist
    return out



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
"""
JournalHint service to display
"Were you looking for a journal reference? Try: <link>"
when the request is a journal reference
"""
from invenio_search.services import SearchService
from invenio.config import (CFG_SITE_URL,
                            CFG_SITE_LANG)
from invenio.base.i18n import gettext_set_language
from invenio.legacy.search_engine import perform_request_search, print_record
from invenio.legacy.webuser import collect_user_info
from urllib import urlencode
import re
from cgi import escape

__plugin_version__ = "Search Service Plugin API 1.0"


class JournalHintService(SearchService):

    """Give hints on how to search the journal reference."""

    re_dot_comma_year = re.compile(r',|\([0-9]{4}\)')
    re_keyword = re.compile(
        r'^((f|fin|find)\s+)|^([a-zA-Z0-9_-]+\:)',
        re.IGNORECASE
    )

    @classmethod
    def seems_a_journal_reference(cls, reference):
        """Quickly check if `name` seems to be a journal reference."""
        # It must not be empty
        if not reference.strip():
            return False
        # There must be a comma or a year in parentesis
        if not cls.re_dot_comma_year.search(reference):
            return False
        # It must not start with a keyword
        if cls.re_keyword.search(reference):
            return False
        return True

    def get_description(self, ln=CFG_SITE_LANG):
        """Return service description."""
        return "Give hints on how to search the journal reference"

    def answer(self, req, user_info, of, cc,
               colls_to_search, p, f, search_units, ln):
        """Answer question given by context.

        Return (relevance, html_string) where relevance is integer
        from 0 to 100 indicating how relevant to the question the
        answer is (see C{CFG_WEBSEARCH_SERVICE_MAX_SERVICE_ANSWER_RELEVANCE}
        for details), and html_string being a formatted answer.
        """
        from invenio.refextract_api import search_from_reference

        _ = gettext_set_language(ln)

        if f or not self.seems_a_journal_reference(p):
            return (0, "")

        (field, pattern) = search_from_reference(p.decode('utf-8'))

        if field is not "journal":
            return (0, "")

        recids = perform_request_search(
            req=req, p=pattern, f=field, cc=cc, c=colls_to_search)

        if not recids:
            return (0, "")

        if len(recids) == 1:
            recid = recids.pop()
            user_info = collect_user_info(req)
            return (100, """\
<p><span class="journalhint">%s</span></p>
<table style="padding: 5px; border: 2px solid #ccc; margin: 20px"><tr><td>
%s
</td></tr></table>""" % (escape(_("Were you looking for this paper?")),
             print_record(recid, ln=ln, user_info=user_info)))

        query = "find rawref \"" + p + "\""
        query_link = CFG_SITE_URL + '/search?' + urlencode({'p': query})
        return (80, '<span class="journalhint">%s</span>' % (
                _("Were you looking for a journal reference? Try: %(x_href)s") %
                {"x_href": '<a href="{0}">{1}</a>'.format(
                 escape(query_link, True), escape(query))}, ))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
"""
WebSearch service to answer based on BibKnowledge KB
"""
from invenio.config import CFG_SITE_LANG
from invenio_search.services import KnowledgeBaseService
from invenio.base.i18n import gettext_set_language

__plugin_version__ = "Search Service Plugin API 1.0"

class FAQKBService(KnowledgeBaseService):

    def get_description(self, ln=CFG_SITE_LANG):
        "Return service description"
        return "Return links of interest based on query"

    def get_label(self, ln=CFG_SITE_LANG):
        "Return label for the list of answers"
        _ = gettext_set_language(ln)
        return _("You might be interested in:")

    def get_kbname(self):
        "Return name of KB to use"
        return "FAQ"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implement AST vistor."""

from invenio_query_parser.ast import (
    AndOp, KeywordOp, OrOp,
    NotOp, Keyword, Value,
    SingleQuotedValue,
    DoubleQuotedValue,
    RegexValue, RangeOp,
    ValueQuery, EmptyQuery
)
from invenio_query_parser.visitor import make_visitor


class Terms(object):

    """Implement visitor to get all given terms."""

    visitor = make_visitor()

    def __init__(self, keywords=None):
        """Initialize list of keywords operators."""
        self.keywords = keywords

    # pylint: disable=W0613,E0102

    @visitor(AndOp)
    def visit(self, node, left, right):
        return left + right

    @visitor(OrOp)
    def visit(self, node, left, right):
        return left + right

    @visitor(NotOp)
    def visit(self, node, op):
        return []

    @visitor(KeywordOp)
    def visit(self, node, left, right):
        return right if left else []

    @visitor(ValueQuery)
    def visit(self, node, op):
        return map(lambda p: p.strip('%'), op)

    @visitor(Keyword)
    def visit(self, node):
        return self.keywords is None or node.value in self.keywords

    @visitor(Value)
    def visit(self, node):
        return [node.value.strip('%')]

    @visitor(SingleQuotedValue)
    def visit(self, node):
        return [node.value]

    @visitor(DoubleQuotedValue)
    def visit(self, node):
        return [node.value]

    @visitor(RegexValue)
    def visit(self, node):
        return [node.value]

    @visitor(RangeOp)
    def visit(self, node, left, right):
        return ["%s->%s" % (left, right)]

    @visitor(EmptyQuery)
    def visit(self, node):
        return []

    # pylint: enable=W0612,E0102



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
"""
WebSearch service to search in collection names
"""
import re
import urllib
from invenio_search.services import ListLinksService
from invenio.base.i18n import gettext_set_language
from invenio.legacy.bibindex.engine_stemmer import stem
from invenio.legacy.dbquery import get_table_update_time
from invenio.config import \
     CFG_WEBSEARCH_COLLECTION_NAMES_SEARCH, \
     CFG_SITE_URL, \
     CFG_SITE_NAME, \
     CFG_SITE_LANG, \
     CFG_CERN_SITE

__plugin_version__ = "Search Service Plugin API 1.0"

whitespace_re = re.compile('\s*')
non_alphanum_chars_only_re = re.compile('\W')

class CollectionNameSearchService(ListLinksService):
    """
    Search collection names
    """
    def get_description(self, ln=CFG_SITE_LANG):
        "Return service description"
        return "Return collections of interest based on query"

    def get_label(self, ln=CFG_SITE_LANG):
        "Return label for the list of answers"
        _ = gettext_set_language(ln)
        return _("Looking for a particular collection? Try:")

    def answer(self, req, user_info, of, cc, colls_to_search, p, f, search_units, ln):
        """
        Answer question given by context.

        Return (relevance, html_string) where relevance is integer
        from 0 to 100 indicating how relevant to the question the
        answer is (see C{CFG_WEBSEARCH_SERVICE_MAX_SERVICE_ANSWER_RELEVANCE} for details) ,
        and html_string being a formatted answer.
        """
        from invenio.legacy.search_engine import \
             get_permitted_restricted_collections, \
             get_coll_i18nname, \
             collection_i18nname_cache, \
             collection_restricted_p
        _ = gettext_set_language(ln)
        # stem search units. remove those with field
        # TODO: search in hosted collection names too
        # TODO: ignore unattached trees
        # TODO: use synonyms
        if f or (CFG_WEBSEARCH_COLLECTION_NAMES_SEARCH < 0) or \
               (CFG_WEBSEARCH_COLLECTION_NAMES_SEARCH == 0 and cc != CFG_SITE_NAME):
            return (0, '')

        words = [stem(unit[1], ln) for unit in search_units if unit[2] in ('', 'collection')] # Stemming

        if not words:
            return (0, '')

        permitted_restricted_collections = get_permitted_restricted_collections(user_info)
        cache = self.get_data_cache()

        matching_collections = {}
        for word in words:
            if CFG_CERN_SITE and word == 'cern':
                # This keyword is useless here...
                continue

            colls = cache.get(word.lower(), [])
            for coll in colls:
                if collection_restricted_p(coll) and \
                       not coll in permitted_restricted_collections:
                    # Skip restricted collection user do not have access
                    continue
                if not matching_collections.has_key(coll):
                    matching_collections[coll] = 0
                matching_collections[coll] += 1


        matching_collections_sorted = sorted(matching_collections.iteritems(), key=lambda (k, v): (v, k), reverse=True)
        if not matching_collections_sorted:
            return (0, '')

        matching_collections_names = [(get_coll_i18nname(coll, ln, False), CFG_SITE_URL + '/collection/' + urllib.quote(coll, safe='') + '?ln=en') \
                                      for coll, score in matching_collections_sorted]

        best_score = matching_collections_sorted[0][1]
        best_coll_words = whitespace_re.split(matching_collections_sorted[0][0])

        relevance = min(100, max(0, (100 * float(2 * best_score) /  float(len(best_coll_words) + len(words)) - 10)))

        if (('submit' in p.lower()) or (_('submit') in p.lower())) and \
               not (('submit' in best_coll_words) or (_('submit') in best_coll_words)):
            # User is probably looking for a submission. Decrease relevance
            relevance = max(0, relevance - 30)

        return (relevance, self.display_answer_helper(matching_collections_names, ln))

    def prepare_data_cache(self):
        """
        "Index" collection names
        """
        from invenio.legacy.search_engine import collection_i18nname_cache
        cache = {}
        words_and_coll = [(' '.join([' '.join([stem(word.lower(), ln) for word in \
                                               whitespace_re.split(non_alphanum_chars_only_re.sub(' ', translation))]) for ln, translation in \
                                     translations.iteritems()]), coll_name) for coll_name, translations in \
                          collection_i18nname_cache.cache.iteritems()]
        for words, coll in words_and_coll:
            for word in whitespace_re.split(words):
                if not word.strip():
                    continue
                if not cache.has_key(word):
                    cache[word] = []
                if not coll in cache[word]:
                    cache[word].append(coll)
        return cache

    def timestamp_verifier(self):
        """
        Return the time at which the data was last updated.  If the
        value returned by the function is newer than the cache, the
        cache will be invalidated.

        @return: string-formatted time '%Y-%m-%d %H:%M:%S'
        """
        return max(get_table_update_time('collectionname'),
                   get_table_update_time('collection_collection'))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implement AST vistor."""

from intbitset import intbitset

from invenio_query_parser.ast import (
    AndOp, DoubleQuotedValue, EmptyQuery,
    GreaterOp, Keyword,
    KeywordOp, NotOp, OrOp,
    RangeOp, RegexValue,
    SingleQuotedValue,
    Value, ValueQuery,
)
from invenio_query_parser.visitor import make_visitor

from ..searchext.engines.native import search_unit


class SearchUnit(object):

    """Implement visitor using ``search_unit`` API."""

    visitor = make_visitor()

    # pylint: disable=W0613,E0102

    @visitor(AndOp)
    def visit(self, node, left, right):
        return left & right

    @visitor(OrOp)
    def visit(self, node, left, right):
        return left | right

    @visitor(NotOp)
    def visit(self, node, op):
        return intbitset(trailing_bits=1) - op

    @visitor(KeywordOp)
    def visit(self, node, left, right):
        if isinstance(right, intbitset):  # second level operator
            left.update(dict(p=right))
        else:
            left.update(right)
        return search_unit(**left)

    @visitor(ValueQuery)
    def visit(self, node, op):
        return search_unit(**op)

    @visitor(GreaterOp)
    def visit(self, node, op):
        op["p"] = "{0}->".format(op["p"])
        return op

    @visitor(Keyword)
    def visit(self, node):
        return dict(f=node.value)

    @visitor(Value)
    def visit(self, node):
        return dict(p=node.value)

    @visitor(SingleQuotedValue)
    def visit(self, node):
        return dict(p=node.value, m='p')

    @visitor(DoubleQuotedValue)
    def visit(self, node):
        return dict(p=node.value, m='e')

    @visitor(RegexValue)
    def visit(self, node):
        return dict(p=node.value, m='r')

    @visitor(RangeOp)
    def visit(self, node, left, right):
        return dict(p="%s->%s" % (left, right))

    @visitor(EmptyQuery)
    def visit(self, node):
        return intbitset(trailing_bits=1)

    # pylint: enable=W0612,E0102



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implement AST vistor."""

import re
from collections import MutableMapping, MutableSequence

import six
from invenio_query_parser.ast import AndOp, DoubleQuotedValue, EmptyQuery, \
    Keyword, KeywordOp, NotOp, OrOp, RangeOp, RegexValue, SingleQuotedValue, \
    Value, ValueQuery
from invenio_query_parser.visitor import make_visitor

from invenio.utils.memoise import memoize


@memoize
def get_field_tags(field, tagtype="marc"):
    """Returns a list of tags for the field code 'field'.

    Example: field='author', output=['100__%','700__%'].
    """
    from invenio_search.models import Field
    return list(Field.get_field_tags(field, tagtype=tagtype))


def match_unit(record, p, f=None, m='a', wl=None):
    """Match record to basic match unit."""
    if record is None:
        return p is None

    if f is not None:
        fields = (get_field_tags(f, 'nonmarc') + [f])
        for field in fields:
            if match_unit(record.get(field), p, f=None, m=m, wl=None):
                return True
        return False

    # compile search value only once for non exact search
    if m != 'e' and isinstance(p, six.string_types):
        p = re.compile(p)

    if isinstance(record, MutableSequence):
        return any([match_unit(field, p, f=f, m=m, wl=wl)
                    for field in record])
    elif isinstance(record, MutableMapping):
        return any([match_unit(field, p, f=f, m=m, wl=wl)
                    for field in record.values()])

    if m == 'e':
        return six.text_type(record) == p
    return p.search(six.text_type(record)) is not None


class MatchUnit(object):

    """Implement visitor using ``match_unit`` API."""

    visitor = make_visitor()

    def __init__(self, record):
        self.record = record

    # pylint: disable=W0613,E0102

    @visitor(AndOp)
    def visit(self, node, left, right):
        return left & right

    @visitor(OrOp)
    def visit(self, node, left, right):
        return left | right

    @visitor(NotOp)
    def visit(self, node, op):
        return not op

    @visitor(KeywordOp)
    def visit(self, node, left, right):
        if isinstance(right, bool):  # second level operator
            left.update(dict(p=right))
        else:
            left.update(right)
        return match_unit(self.record, **left)

    @visitor(ValueQuery)
    def visit(self, node, op):
        return match_unit(self.record, **op)

    @visitor(Keyword)
    def visit(self, node):
        return dict(f=node.value)

    @visitor(Value)
    def visit(self, node):
        return dict(p=node.value)

    @visitor(SingleQuotedValue)
    def visit(self, node):
        return dict(p=node.value, m='p')

    @visitor(DoubleQuotedValue)
    def visit(self, node):
        return dict(p=node.value, m='e')

    @visitor(RegexValue)
    def visit(self, node):
        return dict(p=node.value, m='r')

    @visitor(RangeOp)
    def visit(self, node, left, right):
        return dict(p="%s->%s" % (left, right))

    @visitor(EmptyQuery)
    def visit(self, node):
        return True

    # pylint: enable=W0612,E0102



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Utility functions for search engine."""

import functools

import six

from flask import g
from intbitset import intbitset
from six import iteritems, string_types
from werkzeug.utils import import_string

from invenio.base.globals import cfg
from invenio_collections.cache import (
    get_collection_allchildren,
    restricted_collection_cache,
)


def get_most_popular_field_values(recids, tags, exclude_values=None,
                                  count_repetitive_values=True, split_by=0):
    """Analyze RECIDS and look for TAGS and return most popular values.

    Optionally return the frequency with which they occur sorted according to
    descending frequency.

    If a value is found in EXCLUDE_VALUES, then do not count it.

    If COUNT_REPETITIVE_VALUES is True, then we count every occurrence
    of value in the tags.  If False, then we count the value only once
    regardless of the number of times it may appear in a record.
    (But, if the same value occurs in another record, we count it, of
    course.)

    Example:

    .. code-block:: python

        >>> get_most_popular_field_values(range(11,20), '980__a')
        [('PREPRINT', 10), ('THESIS', 7), ...]
        >>> get_most_popular_field_values(range(11,20), ('100__a', '700__a'))
        [('Ellis, J', 10), ('Ellis, N', 7), ...]
        >>> get_most_popular_field_values(range(11,20), ('100__a', '700__a'),
        ... ('Ellis, J'))
        [('Ellis, N', 7), ...]

    :return: list of tuples containing tag and its frequency
    """
    from invenio.legacy.bibrecord import get_fieldvalues

    valuefreqdict = {}
    # sanity check:
    if not exclude_values:
        exclude_values = []
    if isinstance(tags, string_types):
        tags = (tags,)
    # find values to count:
    vals_to_count = []
    displaytmp = {}
    if count_repetitive_values:
        # counting technique A: can look up many records at once: (very fast)
        for tag in tags:
            vals_to_count.extend(get_fieldvalues(recids, tag, sort=False,
                                                 split_by=split_by))
    else:
        # counting technique B: must count record-by-record: (slow)
        for recid in recids:
            vals_in_rec = []
            for tag in tags:
                for val in get_fieldvalues(recid, tag, False):
                    vals_in_rec.append(val)
            # do not count repetitive values within this record
            # (even across various tags, so need to unify again):
            dtmp = {}
            for val in vals_in_rec:
                dtmp[val.lower()] = 1
                displaytmp[val.lower()] = val
            vals_in_rec = dtmp.keys()
            vals_to_count.extend(vals_in_rec)
    # are we to exclude some of found values?
    for val in vals_to_count:
        if val not in exclude_values:
            if val in valuefreqdict:
                valuefreqdict[val] += 1
            else:
                valuefreqdict[val] = 1
    # sort by descending frequency of values:
    f = []   # frequencies
    n = []   # original names
    ln = []  # lowercased names
    # build lists within one iteration
    for (val, freq) in iteritems(valuefreqdict):
        f.append(-1 * freq)
        if val in displaytmp:
            n.append(displaytmp[val])
        else:
            n.append(val)
        ln.append(val.lower())
    # sort by frequency (desc) and then by lowercased name.
    try:
        import numpy
        indices = numpy.lexsort([ln, f])
    except ImportError:
        def _cmp(a, b):
            if f[a] == f[b]:
                return ln[a] - ln[b]
            else:
                return f[a] - f[b]
        indices = sorted(
            range(min(len(ln), len(f))),
            cmp=_cmp
        )
    return [(n[i], -1 * f[i]) for i in indices]


def get_permitted_restricted_collections(user_info,
                                         recreate_cache_if_needed=True):
    """Return a list of restricted collection with user is authorization."""
    from invenio.modules.access.engine import acc_authorize_action

    if recreate_cache_if_needed:
        restricted_collection_cache.recreate_cache_if_needed()
    ret = []

    auths = acc_authorize_action(
        user_info,
        'viewrestrcoll',
        batch_args=True,
        collection=restricted_collection_cache.cache
    )

    for collection, auth in zip(restricted_collection_cache.cache, auths):
        if auth[0] == 0:
            ret.append(collection)
    return ret


def g_memoise(method=None, key=None):
    """Memoise method results on application context."""
    if method is None:
        return functools.partial(g_memoise, key=key)

    key = key or method.__name__

    @functools.wraps(method)
    def decorator(*args, **kwargs):
        results = getattr(g, key, None)
        if results is None:
            results = method(*args, **kwargs)
            setattr(g, key, results)
        return results
    return decorator


@g_memoise
def query_enhancers():
    """Return list of query enhancers."""
    functions = []
    for enhancer in cfg['SEARCH_QUERY_ENHANCERS']:
        if isinstance(enhancer, six.string_types):
            enhancer = import_string(enhancer)
            functions.append(enhancer)
    return functions


@g_memoise
def parser():
    """Return search query parser."""
    query_parser = cfg['SEARCH_QUERY_PARSER']
    if isinstance(query_parser, six.string_types):
        query_parser = import_string(query_parser)
    return query_parser


@g_memoise
def query_walkers():
    """Return query walker instances."""
    return [
        import_string(walker)() if isinstance(walker, six.string_types)
        else walker() for walker in cfg['SEARCH_QUERY_WALKERS']
    ]


@g_memoise
def search_walkers():
    """Return in search walker instances."""
    return [
        import_string(walker)() if isinstance(walker, six.string_types)
        else walker() for walker in cfg['SEARCH_WALKERS']
    ]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Test policies on restricted and permitted collections."""

from invenio_search.enhancers.collection_filter import \
    create_collection_query
from invenio.testsuite import InvenioTestCase, make_test_suite, run_test_suite

from invenio_query_parser.ast import AndOp, NotOp, OrOp


class MockedRecord(object):

    _all_records = []

    def __init__(self, collections):
        """
        :param: collections: A list of the collections that the record belongs
        """
        self.collections = collections
        MockedRecord._all_records.append(self)

    def __str__(self):
        return str(self.collections)

    def belongsTo(self, collection):
        return collection in self.collections

    def belongs_to_c_expr(self, c_expr):
        """
        :param c_expr: A boolean invenio_query_parser expr
                                eg. AndOp("a", "b")
                                or a collection name
        """
        if type(c_expr) is str:
            return self.belongsTo(c_expr)

        class_name = c_expr.__class__.__name__
        if class_name == "AndOp":
            return self.belongs_to_c_expr(c_expr.right) and \
                self.belongs_to_c_expr(c_expr.left)
        elif class_name == "OrOp":
            return self.belongs_to_c_expr(c_expr.right) or \
                self.belongs_to_c_expr(c_expr.left)
        elif class_name == "NotOp":
            return not self.belongs_to_c_expr(c_expr.op)


class RecordsRegistry(object):

    """Keep track of all the records that are created."""

    def __init__(self):
        self.all_records = []

    def create_record(self, collections):
        r = MockedRecord(collections)
        self.all_records.append(r)
        return r

    def filter_collections(self, c_expr):
        return filter(lambda x: x.belongs_to_c_expr(c_expr), self.all_records)


def create_query(restricted_collections, permitted_restricted_collections,
                 current_col, policy):
    return create_collection_query(restricted_collections,
                                   permitted_restricted_collections,
                                   current_col, policy,
                                   formatter=lambda x: x)


def get_records_to_show(restricted_collections,
                        permitted_restricted_colls, current_col,
                        record_registry, policy):
    collection_query = create_query(restricted_collections,
                                    permitted_restricted_colls,
                                    current_col, policy)

    return record_registry.filter_collections(collection_query)


class TestCollectionsFilterEnhancher(InvenioTestCase):

    def setUp(self):

        self.record_reg = RecordsRegistry()
        self.r1 = self.record_reg.create_record(["a", "cc"])
        self.r2 = self.record_reg.create_record(["a", "b", "cc"])
        self.r3 = self.record_reg.create_record(["b", "cc"])
        self.r4 = self.record_reg.create_record(["b", "c", "cc"])
        self.r5 = self.record_reg.create_record(["c", "cc"])
        self.r6 = self.record_reg.create_record(["a", "cc"])
        self.r7 = self.record_reg.create_record(["a", "d", "cc"])
        self.r8 = self.record_reg.create_record(["d", "cc"])
        self.r9 = self.record_reg.create_record(["d", "c", "cc"])
        self.r10 = self.record_reg.create_record(["c", "cc"])
        self.r11 = self.record_reg.create_record(["cc"])

    def test_simple(self):
        self.assertEqual(self.record_reg.filter_collections("a"),
                         [self.r1, self.r2, self.r6, self.r7])

    def test_col_expr(self):
        """Test record filtering"""
        self.assertEqual(self.r1.belongs_to_c_expr("a"), True)
        self.assertEqual(self.r1.belongs_to_c_expr("b"), False)
        self.assertEqual(self.r1.belongs_to_c_expr(AndOp("a", "b")), False)
        self.assertEqual(self.r1.belongs_to_c_expr(OrOp("a", "b")), True)
        self.assertEqual(self.r1.belongs_to_c_expr(AndOp("d", OrOp("a", "b"))),
                         False)
        self.assertEqual(self.r1.belongs_to_c_expr(AndOp("a", NotOp("b"))),
                         True)

    def test_policy_else(self):
        self.assertEqual(get_records_to_show(["a", "b", "c", "d"], ["a"],
                                             "cc", self.record_reg, "ELSE"),
                         [self.r1, self.r6, self.r11])

    def test_policy_any(self):
        self.assertEqual(get_records_to_show(["a", "b", "c", "d"], ["a"],
                                             "cc", self.record_reg, "ANY"),
                         [self.r1, self.r2, self.r6, self.r7, self.r11])

    def test_none(self):
        self.assertEqual(create_query([], [], 'cc', 'ANY'), 'cc')

    def test_no_restrict(self):
        self.assertEqual(create_query(['a', 'b'], ['a', 'b'], 'cc', 'ANY'),
                         'cc')

TEST_SUITE = make_test_suite(TestCollectionsFilterEnhancher)

if __name__ == "__main__":
    run_test_suite(TEST_SUITE)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implement AST vistor."""

from invenio_query_parser.ast import (
    AndOp,
    DoubleQuotedValue,
    EmptyQuery,
    Keyword,
    KeywordOp,
    NotOp,
    OrOp,
    RangeOp,
    RegexValue,
    SingleQuotedValue,
    Value,
    ValueQuery,
)

from invenio_query_parser.visitor import make_visitor


class FacetsVisitor(object):

    """Implement visitor to extract all facets filters."""

    visitor = make_visitor()

    @staticmethod
    def jsonable(parsedFacets):
        """Convert a visited query result to a structure which can be jsonified.

        :param parsedFacets: a visited query result.
        """
        result = {}
        # sets cannot be converted to json. We need to convert them to lists.
        for facet_name in parsedFacets:
            result[facet_name] = {
                'inc': list(parsedFacets[facet_name]['inc']),
                'exc': list(parsedFacets[facet_name]['exc']),
            }
        return result

    # pylint: disable=W0613,E0102,F999,D102

    def _merge_facets(self, left, right):
        """merge faceting for an AND or OR operator.

        :param left: left child node faceting
        :param right: right child node faceting
        """
        for k in right:
            if k in left:
                inc = left[k]['inc'].union(right[k]['inc'])
                exc = left[k]['exc'].union(right[k]['exc'])
                # Don't mark as included or excluded if only partially
                # included/excluded
                left[k] = {
                    'inc': inc.difference(exc),
                    'exc': exc.difference(inc),
                }
            else:
                left[k] = right[k]
        return left

    def _invert_facets(self, facets):
        """invert facet filters included <-> excluded.

        :param facets: facet filters
        """
        for k in facets:
            facets[k] = {
                'inc': facets[k]['exc'],
                'exc': facets[k]['inc'],
            }
        return facets

    @visitor(AndOp)
    def visit(self, node, left, right):
        return self._merge_facets(left, right)

    @visitor(OrOp)
    def visit(self, node, left, right):
        return self._merge_facets(left, right)

    @visitor(NotOp)
    def visit(self, node, op):
        return self._invert_facets(op)

    @visitor(KeywordOp)
    def visit(self, node, left, right):
        return {
            node.left.value: {
                'inc': set([node.right.value]),
                'exc': set()
            }
        }

    @visitor(ValueQuery)
    def visit(self, node, op):
        return {}

    @visitor(Keyword)
    def visit(self, node):
        return {}

    @visitor(Value)
    def visit(self, node):
        return {}

    @visitor(SingleQuotedValue)
    def visit(self, node):
        return {}

    @visitor(DoubleQuotedValue)
    def visit(self, node):
        return {}

    @visitor(RegexValue)
    def visit(self, node):
        return {}

    @visitor(RangeOp)
    def visit(self, node, left, right):
        return {}

    @visitor(EmptyQuery)
    def visit(self, node):
        return {}
    # pylint: enable=W0612,E0102,F999,D102



# This file is part of Invenio.
# Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implement washers for search arguments."""

import re
import time

from invenio.base.globals import cfg
from invenio.utils.datastructures import LazyDict
from invenio.utils.text import wash_for_utf8

re_pattern_wildcards_after_spaces = re.compile(r'(\s)[\*\%]+')
re_pattern_single_quotes = re.compile("'(.*?)'")
re_pattern_double_quotes = re.compile("\"(.*?)\"")
re_pattern_parens_quotes = re.compile(r'[\'\"]{1}[^\'\"]*(\([^\'\"]*\))[^\'\"]'
                                      '*[\'\"]{1}')
re_pattern_regexp_quotes = re.compile(r"\/(.*?)\/")
re_pattern_short_words = re.compile(r'([\s\"]\w{1,3})[\*\%]+')
re_pattern_space = re.compile("__SPACE__")
re_pattern_today = re.compile(r"\$TODAY\$")


def get_search_results_default_urlargd():
    """Return default config for search results arguments."""
    return {
        'cc': (str, cfg['CFG_SITE_NAME']),
        'c': (list, []),
        'p': (str, ""), 'f': (str, ""),
        'rg': (int, cfg['CFG_WEBSEARCH_DEF_RECORDS_IN_GROUPS']),
        'sf': (str, ""),
        'so': (str, "d"),
        'sp': (str, ""),
        'rm': (str, ""),
        'of': (str, "hb"),
        'ot': (list, []),
        'em': (str, ""),
        'aas': (int, cfg['CFG_WEBSEARCH_DEFAULT_SEARCH_INTERFACE']),
        'as': (int, cfg['CFG_WEBSEARCH_DEFAULT_SEARCH_INTERFACE']),
        'p1': (str, ""), 'f1': (str, ""), 'm1': (str, ""), 'op1': (str, ""),
        'p2': (str, ""), 'f2': (str, ""), 'm2': (str, ""), 'op2': (str, ""),
        'p3': (str, ""), 'f3': (str, ""), 'm3': (str, ""),
        'sc': (int, 0),
        'jrec': (int, 0),
        'recid': (int, -1), 'recidb': (int, -1), 'sysno': (str, ""),
        'id': (int, -1), 'idb': (int, -1), 'sysnb': (str, ""),
        'action': (str, "search"),
        'action_search': (str, ""),
        'action_browse': (str, ""),
        'd1': (str, ""),
        'd1y': (int, 0), 'd1m': (int, 0), 'd1d': (int, 0),
        'd2': (str, ""),
        'd2y': (int, 0), 'd2m': (int, 0), 'd2d': (int, 0),
        'dt': (str, ""),
        'ap': (int, 1),
        'verbose': (int, 0),
        'ec': (list, []),
        'wl': (int, cfg['CFG_WEBSEARCH_WILDCARD_LIMIT']),
        }

search_results_default_urlargd = LazyDict(get_search_results_default_urlargd)


def wash_pattern(p):
    """Wash pattern passed in URL.

    Check for sanity of the wildcard by removing wildcards if they are appended
    to extremely short words (1-3 letters).
    """
    # strip accents:
    # p = strip_accents(p) # FIXME: when available, strip accents all the time
    # add leading/trailing whitespace for the two following wildcard-sanity
    # checking regexps:
    p = " " + p + " "
    # replace spaces within quotes by __SPACE__ temporarily:
    p = re_pattern_single_quotes.sub(
        lambda x: "'" + x.group(1).replace(' ', '__SPACE__') + "'", p)
    p = re_pattern_double_quotes.sub(
        lambda x: "\""+x.group(1).replace(' ', '__SPACE__') + "\"", p)
    p = re_pattern_regexp_quotes.sub(
        lambda x: "/" + x.group(1).replace(' ', '__SPACE__') + "/", p)
    # get rid of unquoted wildcards after spaces:
    p = re_pattern_wildcards_after_spaces.sub("\\1", p)
    # get rid of extremely short words (1-3 letters with wildcards):
    # p = re_pattern_short_words.sub("\\1", p)
    # replace back __SPACE__ by spaces:
    p = re_pattern_space.sub(" ", p)
    # replace special terms:
    p = re_pattern_today.sub(time.strftime("%Y-%m-%d", time.localtime()), p)
    # remove unnecessary whitespace:
    p = p.strip()
    # remove potentially wrong UTF-8 characters:
    p = wash_for_utf8(p)
    return p


def wash_output_format(ouput_format):
    """Wash output format FORMAT.

    Currently only prevents input like 'of=9' for backwards-compatible format
    that prints certain fields only (for this task, 'of=tm' is preferred).
    """
    if str(ouput_format[0:3]).isdigit() and len(ouput_format) != 6:
        # asked to print MARC tags, but not enough digits,
        # so let's switch back to HTML brief default
        return 'hb'
    return ouput_format


def wash_field(f):
    """Wash field passed by URL."""
    if f:
        # get rid of unnecessary whitespace and make it lowercase
        # (e.g. Author -> author) to better suit iPhone etc input
        # mode:
        f = f.strip().lower()
    # wash legacy 'f' field names, e.g. replace 'wau' or `au' by
    # 'author', if applicable:
    return cfg.get('CFG_WEBSEARCH_FIELDS_CONVERT', {}).get(f, f)


def wash_dates(d1="", d1y=0, d1m=0, d1d=0, d2="", d2y=0, d2m=0, d2d=0):
    """Wash date passed by URL.

    Take user-submitted date arguments D1 (full datetime string) or
    (D1Y, D1M, D1Y) year, month, day tuple and D2 or (D2Y, D2M, D2Y)
    and return (YYY1-M1-D2 H1:M1:S2, YYY2-M2-D2 H2:M2:S2) datetime
    strings in the YYYY-MM-DD HH:MM:SS format suitable for time
    restricted searching.

    Note that when both D1 and (D1Y, D1M, D1D) parameters are present,
    the precedence goes to D1.  Ditto for D2*.

    Note that when (D1Y, D1M, D1D) are taken into account, some values
    may be missing and are completed e.g. to 01 or 12 according to
    whether it is the starting or the ending date.
    """
    datetext1, datetext2 = "", ""
    # sanity checking:
    if d1 == "" and d1y == 0 and d1m == 0 and d1d == 0 and \
            d2 == "" and d2y == 0 and d2m == 0 and d2d == 0:
        return ("", "")  # nothing selected, so return empty values
    # wash first (starting) date:
    if d1:
        # full datetime string takes precedence:
        datetext1 = d1
    else:
        # okay, first date passed as (year,month,day):
        if d1y:
            datetext1 += "%04d" % d1y
        else:
            datetext1 += "0000"
        if d1m:
            datetext1 += "-%02d" % d1m
        else:
            datetext1 += "-01"
        if d1d:
            datetext1 += "-%02d" % d1d
        else:
            datetext1 += "-01"
        datetext1 += " 00:00:00"
    # wash second (ending) date:
    if d2:
        # full datetime string takes precedence:
        datetext2 = d2
    else:
        # okay, second date passed as (year,month,day):
        if d2y:
            datetext2 += "%04d" % d2y
        else:
            datetext2 += "9999"
        if d2m:
            datetext2 += "-%02d" % d2m
        else:
            datetext2 += "-12"
        if d2d:
            datetext2 += "-%02d" % d2d
        else:
            datetext2 += "-31"
            # NOTE: perhaps we should add max(datenumber) in given month,
            # but for our quering it's not needed, 31 will always do.
        datetext2 += " 00:00:00"
    # okay, return constructed YYYY-MM-DD HH:MM:SS datetexts:
    return (datetext1, datetext2)


def wash_search_urlargd(form):
    """
    Create canonical search arguments from those passed via web form.
    """
    from invenio.ext.legacy.handler import wash_urlargd
    argd = wash_urlargd(form, search_results_default_urlargd)
    if 'as' in argd:
        argd['aas'] = argd['as']
        del argd['as']
    if argd.get('aas', cfg['CFG_WEBSEARCH_DEFAULT_SEARCH_INTERFACE']) \
            not in cfg['CFG_WEBSEARCH_ENABLED_SEARCH_INTERFACES']:
        argd['aas'] = cfg['CFG_WEBSEARCH_DEFAULT_SEARCH_INTERFACE']

    # Sometimes, users pass ot=245,700 instead of
    # ot=245&ot=700. Normalize that.
    ots = []
    for ot in argd['ot']:
        ots += ot.split(',')
    argd['ot'] = ots

    # We can either get the mode of function as
    # action=<browse|search>, or by setting action_browse or
    # action_search.
    if argd['action_browse']:
        argd['action'] = 'browse'
    elif argd['action_search']:
        argd['action'] = 'search'
    else:
        if argd['action'] not in ('browse', 'search'):
            argd['action'] = 'search'

    del argd['action_browse']
    del argd['action_search']

    if argd['em'] != "":
        argd['em'] = argd['em'].split(",")

    return argd



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012, 2014 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.import re

"""Provide the infrastructure to load/query search services."""

import re
from invenio.config import CFG_SITE_LANG
from invenio.modules.knowledge.api import get_kb_mappings
from invenio.legacy.miscutil.data_cacher import DataCacher
from invenio.legacy.bibindex.engine_stemmer import stem
from invenio.legacy.dbquery import get_table_update_time
from invenio.base.i18n import gettext_set_language
from invenio.legacy import template

from . import registry

CFG_WEBSEARCH_SERVICE_MAX_SERVICE_ANSWER_RELEVANCE = 100  # from 0 to 100
"""
The following scale is should be considered by services:

- 100: the service thinks it has found the only answer to the question.
  No other service can find an answer, and no other service will be displayed.

- 90: The service thinks it has found the good answer, probably the one
  expected by the user. However some other services with really good relevance
  might still run.

- 80: The service thinks it has found a good answer

- 50: The service thinks the answer might be useful to the user in some cases

- 30: The service found some answer, which might help the user in some cases

- 20: The service found some answer, but it is unlikely that it will help the
  user.

- 10: The service could not give an answer to the question

- 0: The service could not understand the question
"""

CFG_WEBSEARCH_SERVICE_MAX_RELEVANCE_DIFFERENCE = 30
"""
What is the max distance from the most relevant answer do we still display the
answer from a service? For eg, with
CFG_WEBSEARCH_SERVICE_MAX_RELEVANCE_DIFFERENCE==30 and the most relevant
service returning a relevance of 80, we would still display a service relevant
at 50 or above, but not a service with a relevance below 50
"""

CFG_WEBSEARCH_SERVICE_MIN_RELEVANCE_TO_DISPLAY = 21
"""Consider that a service is not relevant enough to be displayed."""

CFG_WEBSEARCH_SERVICE_MAX_NB_SERVICE_DISPLAY = 2
"""Max number of answers/services to display to the user."""

CFG_WEBSEARCH_MAX_SEARCH_COLL_RESULTS_TO_PRINT = 4
"""
Maximum number of matching collections names to be displayed on the
search results page. All the rest of the collections will be hidden
by a "See more collections" link.
"""


# Base class
class SearchService:

    """Abstract base class for search services.

    New services should subclass this class, and:
      - Override L{get_decription} to specify service description
      - Override L{prepare_data_cache} to prepare some cache needed for answers
      - Override L{timestamp_verifier} to indicate if cache must be refreshed
      - Override L{answer} to return answer (score, html_string)

    Services that inherits directly from this class are fully
    responsible for displaying the output. See also
    L{ListLinksService} and L{KnowledgeBaseService}
    """

    cache = None

    def get_description(self, ln=CFG_SITE_LANG):
        """Return service description.

        :rtype: string
        """
        return ""

    def prepare_data_cache(self):
        """Helper function to cache some data.

        If you need to pre-process some data and fill a cache,
        overrides this method and return the data to be cached. You
        can invalidate this data by overridding method
        L{timestamp_verifier} and retrieve the data from
        L{get_data_cache}.
        """
        return None

    def timestamp_verifier(self):
        """Return the time at which your data was last updated.

        If the value returned by the function is newer than the cache, the
        cache will be invalidated.

        :return: string-formatted time '%Y-%m-%d %H:%M:%S'
        """
        return lambda x: 0

    def answer(self, req, user_info, of, cc, colls_to_search, p, f,
               search_units, ln):
        """Answer question given by context.

        Returns (relevance, html_string) where relevance is integer
        from 0 to 100 indicating how relevant to the question the
        answer is (see L{CFG_WEBSEARCH_SERVICE_MAX_SERVICE_ANSWER_RELEVANCE}
        for details), and html_string being an HTML-formatted answer.

        @rtype: tuple(int, string)
        """
        return (0, "")

    def get_data_cache(self, recreate_cache_if_needed=True):
        """Return always up-to-date data from cache.

        The returned value depends on the what has been stored with
        :meth:`~SearchService.prepare_data_cache`.

        :param recreate_cache_if_needed: if True, force refreshing data cache.
        """
        try:
            self.__class__.cache.is_ok_p
        except Exception:
            self.__class__.cache = DataCacher(self.prepare_data_cache,
                                              self.timestamp_verifier)
        if recreate_cache_if_needed:
            self.__class__.cache.recreate_cache_if_needed()

        return self.__class__.cache.cache


# List display service
class ListLinksService(SearchService):
    """
    Abstract class for services that displays an inline list of links
    as answers.

    In order to provide a service that returns a list of answers in a
    standardized way, subclass this class and:
      - Override L{get_decription} to specify service description
      - Override L{prepare_data_cache} to prepare some cache needed for answers
      - Override L{timestamp_verifier} to indicate if cache must be refreshed
      - Override L{answer} to return answer (score, html_string)
      - Override L{get_label} to specify service output label

    Services that inherits this class can make use of
    L{self.display_answer_helper} function to format the response in
    L{answer} function.

    See also L{KnowledgeBaseService}.
    """
    def get_label(self, ln=CFG_SITE_LANG):
        """
        Return label displayed next to the service responses.

        @rtype: string
        """
        return ""

    def display_answer_helper(self, labels_and_urls, ln=CFG_SITE_LANG):
        """Display HTML to return an answer as HTML.

        Use this function to process your list of response and return
        an HTML-formatted string in your L{answer} function.

        :param labels_and_urls: list of answers in the form (label, URL)
        :param ln: the language in which is used to answer
        :return: HTML-formatted string
        """
        websearch_templates = template.load('websearch')
        label = self.get_label(ln=ln)
        return websearch_templates.tmpl_print_service_list_links(
            label, labels_and_urls, ln=ln)


# KnowledgeBase type service
class KnowledgeBaseService(ListLinksService):

    """Abstract class for Knowledge Base based services.

    In order to provide a service that returns a list of answers in a
    standardized way using a Knowledge Basee, subclass this class and:
      - Override L{get_decription} to specify service description
      - Override L{get_label} to specify service output label
      - Override L{get_kbname} to specify knowledge base to consider

    The answer is retrieved from the knowledge base defined by
    L{get_kbname}, and must map a list of whitespace-separated
    words to an answer in the form "label|url". See for eg. KB 'FAQ'.
    """

    def get_kbname(self):
        """Return name of the knowledge base to use for answering.

        The knowedge base should be defined in BibKnowledge, and must
        map a list of whitespace-separated words to an answer in the
        form C{label|url}.

        @rtype: string
        """
        return ""

    def answer(self, req, user_info, of, cc, colls_to_search, p, f,
               search_units, ln):
        """Answer question given by context, using knowledge base.

        Return (relevance, html_string) where relevance is integer
        from 0 to 100 indicating how relevant to the question the
        answer is (see L{CFG_WEBSEARCH_SERVICE_MAX_SERVICE_ANSWER_RELEVANCE}
        for details), and html_string being a formatted answer.
        """
        _ = gettext_set_language(ln)
        # words = [stem(unit[1], ln) for unit in search_units if unit[2] == '']
        words = [stem(unit[1].lower(), CFG_SITE_LANG) for unit in search_units
                 if unit[2] == '']
        cache = self.get_data_cache()

        matching_values = {}
        for word in words:
            res = cache.get(word, [])
            for keyword in res:
                if keyword not in matching_values:
                    matching_values[keyword] = 1
                else:
                    matching_values[keyword] += 1

        # order matching values per score
        matching_values_sorted = sorted(matching_values.iteritems(),
                                        key=lambda (k, v): (v, k),
                                        reverse=True)

        if not matching_values_sorted:
            return (0, '')

        best_score = matching_values_sorted[0][1]

        # Compute relevance. How many words from query did match
        relevance = min(100, max(
            0, (100 * float(best_score) / len(
                [word for word in words if len(word) > 3]
            )) - 10))
        labels_and_links = [m.split("|", 1) for m in matching_values.keys()]
        translated_labels_and_links = [(_(label), url)
                                       for label, url in labels_and_links]

        return (relevance, self.display_answer_helper(
            translated_labels_and_links, ln))

    def prepare_data_cache(self):
        """*Index* knowledge base and cache it."""
        cache = {}
        for mapping in get_kb_mappings(self.get_kbname()):
            key = mapping['key']
            value = mapping['value']
            words = clean_and_split_words_and_stem(key, CFG_SITE_LANG,
                                                   stem_p=True)
            for word in words:
                if word not in cache:
                    cache[word] = []
                if value not in cache[word]:
                    cache[word].append(value)
        return cache

    def timestamp_verifier(self):
        """Return the time at which the data was last updated.

        If the value returned by the function is newer than the cache, the
        cache will be invalidated.

        :return: string-formatted time ``'%Y-%m-%d %H:%M:%S'``
        """
        # This is an approximation...
        return get_table_update_time('knwKBRVAL')

re_split_words_pattern = re.compile('\s*')
re_non_alphanum_only = re.compile('\W')


def clean_and_split_words_and_stem(string, ln=CFG_SITE_LANG, stem_p=True):
    """Split and stemp words in a string.

    :param ln: language to consider for stemming
    :param stem_p: if True, also stem the word according to ``ln``
    :return: list of (stemmed) word.
    """
    alphanum_string = re_non_alphanum_only.sub(" ", string).lower()
    words = re_split_words_pattern.split(alphanum_string)
    if stem_p:
        # lowering must be done after stemming
        words = [stem(word, ln) for word in words]

    return words


def get_answers(req, user_info, of, cc, colls_to_search, p, f, ln):
    """Return answers from all registered search services."""
    if p:
        from invenio.legacy.search_engine import create_basic_search_units
        search_units = create_basic_search_units(req, p, f)
    else:
        search_units = []

    def search_service_answers():
        for search_service in registry.services:
            yield search_service.answer(req, user_info, of, cc,
                                        colls_to_search, p, f,
                                        search_units, ln)

    nb_answers = 0
    best_relevance = None

    for answer_relevance, answer_html in sorted(
            search_service_answers(), reverse=True):
        nb_answers += 1
        if best_relevance is None:
            best_relevance = answer_relevance
        if best_relevance <= CFG_WEBSEARCH_SERVICE_MIN_RELEVANCE_TO_DISPLAY:
            # The answer is not relevant enough
            break
        if nb_answers > CFG_WEBSEARCH_SERVICE_MAX_NB_SERVICE_DISPLAY:
            # We have reached the max number of service to display
            break
        if best_relevance - answer_relevance > \
                CFG_WEBSEARCH_SERVICE_MAX_RELEVANCE_DIFFERENCE:
            # The service gave an answer that is way less good than previous
            # ones.
            break
        yield answer_relevance, answer_html

        if answer_relevance == \
                CFG_WEBSEARCH_SERVICE_MAX_SERVICE_ANSWER_RELEVANCE:
            # The service assumes it has given the definitive answer
            break



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Test facet filter enhancer."""

from __future__ import unicode_literals

from invenio.testsuite import InvenioTestCase, make_test_suite, run_test_suite

from invenio_query_parser.ast import (
    AndOp, DoubleQuotedValue, EmptyQuery, Keyword, KeywordOp, NotOp, OrOp
)


class TestFacetFilterEnhancer(InvenioTestCase):

    def test_grouped_facets(self):
        from invenio_search.enhancers.facet_filter import \
            get_groupped_facets
        facet_list = [['+', 'foo', 'bar'], ['+', 'foo', 'baz'],
                      ['-', 'bar', 'boo'], ['-', 'bar', 'bo']]

        grouped_facets = {
            '+': {
                'foo': ['bar', 'baz']
            },
            '-': {
                'bar': ['boo', 'bo']
            }
        }
        self.assertEqual(grouped_facets,
                         get_groupped_facets(facet_list))

    def test_format_tree_nodes(self):
        from invenio_search.enhancers.facet_filter import \
            format_facet_tree_nodes
        grouped_facets = {
            '+': {
                'foo': ['bar', 'baz']
            },
            '-': {
                'bar': ['boo', 'bo']
            }
        }
        facet_kws = ['foo', 'bar']
        tree = format_facet_tree_nodes(EmptyQuery(''), grouped_facets,
                                       facet_kws)
        correct = AndOp(
            AndOp(
                EmptyQuery(''),
                OrOp(KeywordOp(Keyword('foo'), DoubleQuotedValue('bar')),
                     KeywordOp(Keyword('foo'), DoubleQuotedValue('baz')))
            ),
            NotOp(OrOp(
                KeywordOp(Keyword('bar'), DoubleQuotedValue('boo')),
                KeywordOp(Keyword('bar'), DoubleQuotedValue('bo')))
            )
        )
        self.assertEqual(correct, tree)

    def test_format_kw_not_in_list(self):
        from invenio_search.enhancers.facet_filter import \
            format_facet_tree_nodes
        grouped_facets = {
            '+': {
                'foo': ['bar', 'baz']
            },
            '-': {
                'bar': ['boo', 'bo']
            }
        }
        facet_kws = ['foo']
        tree = format_facet_tree_nodes(EmptyQuery(''), grouped_facets,
                                       facet_kws)
        correct = AndOp(
            EmptyQuery(''),
            OrOp(KeywordOp(Keyword('foo'), DoubleQuotedValue('bar')),
                 KeywordOp(Keyword('foo'), DoubleQuotedValue('baz')))
        )
        self.assertEqual(correct, tree)

    def test_missing_part(self):
        from invenio_search.enhancers.facet_filter import \
            format_facet_tree_nodes
        grouped_facets = {'-': {'bar': ['boo', 'bo']}}
        facet_kws = ['bar']
        tree = format_facet_tree_nodes(EmptyQuery(''), grouped_facets,
                                       facet_kws)
        correct = AndOp(
            EmptyQuery(''),
            NotOp(OrOp(KeywordOp(Keyword('bar'), DoubleQuotedValue('boo')),
                       KeywordOp(Keyword('bar'), DoubleQuotedValue('bo')))))
        self.assertEqual(correct, tree)

    def test_none(self):
        from invenio_search.enhancers.facet_filter import \
            format_facet_tree_nodes
        self.assertEqual(format_facet_tree_nodes(None, {}, None), None)

TEST_SUITE = make_test_suite(TestFacetFilterEnhancer)

if __name__ == "__main__":
    run_test_suite(TEST_SUITE)
